/*********************************************************************
** AreaBuilder is a program designed to build maps for MUDS like    **
** Anatolia.                                                        **
** Copyright (C) 2003 Pierre Chatelier                              **
**                                                                  **
** This program is free software; you can redistribute              **
** it and/or modify it under the terms of the GNU                   **
** General Public License as published by the                       **
** Free Software Foundation, version 2.                             **
**                                                                  **
** This program is distributed in the hope that it will be          **
** useful, but WITHOUT ANY WARRANTY; without even the implied       **
** warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. **
** See the GNU General Public License for more details.             **
**                                                                  **
** You should have received a copy of the                           **
** GNU General Public License along with this program; if not,      **
** write to the Free Software Foundation,                           **
** Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA      **
*********************************************************************/
package areabuilder;

import java.awt.*;
import java.awt.geom.*; //Point2D, segment...
import javax.swing.*;
import java.util.*; //for LinkedList
import java.io.*;   //for archivation

class Room extends LinkingAndSelectableNodeObject implements Serializable
{
  static public final byte DOORNORTH = AnatoliaConsts.NORTH;
  static public final byte DOOREAST  = AnatoliaConsts.EAST;
  static public final byte DOORSOUTH = AnatoliaConsts.SOUTH;
  static public final byte DOORWEST  = AnatoliaConsts.WEST;
  static public final byte DOORUP    = AnatoliaConsts.UP;
  static public final byte DOORDOWN  = AnatoliaConsts.DOWN;
  static public final byte DOORNONE  = 6;

  static public final int SIZE = 30;

  AnatoliaRoom anatoliaRoom = null;
  boolean autoVnum = true;

  private int x;
  private int y;

  private int sizeOnScreen;
  private int xOnScreen;
  private int yOnScreen;
  private int candidateXOnScreen;
  private int candidateYOnScreen;

  //array indexed on the 6 possibled doors.
  //Contains pairs <room,wire>, meaning connection "here --wire--> room"
  Pair[] connectingToRooms = null; //<room,wire>
  LinkedList mayComeFromRooms = null; //<room, door>
//  private

  int layer = 0; //to store on which floor the room is

  private float scale;

  AnatoliaCollection myCollection = null;
  AreaMap areaMap = null;

  public Room(AreaMap areaMap, AnatoliaCollection anatoliaCollection)
  {
      this.myCollection = anatoliaCollection;
      this.areaMap = areaMap;
      this.anatoliaRoom = new AnatoliaRoom(this, anatoliaCollection);
      this.candidateXOnScreen = this.xOnScreen = 0;
      this.candidateYOnScreen = this.yOnScreen = 0;
      this.scale = 1.f;
      this.x = 0;
      this.y = 0;
      this.sizeOnScreen = (int) (SIZE*scale);
      this.selected = false;
      this.connectingToRooms = new Pair[6]; //rooms we can go to trough each of the 6 possible doors
      for (byte door = 1 ; door <= 6; ++door)
      {
         connectingToRooms[door-1] = null;
      }
      mayComeFromRooms = new LinkedList();
  }
  public Room(AreaMap areaMap, int xOnScreen, int yOnScreen, float scale, AnatoliaCollection anatoliaCollection)
  {
      this.myCollection = anatoliaCollection;

      this.areaMap = areaMap;
      this.anatoliaRoom = new AnatoliaRoom(this, myCollection);
      this.candidateXOnScreen = this.xOnScreen = xOnScreen;
      this.candidateYOnScreen = this.yOnScreen = yOnScreen;
      this.scale = scale;
      this.x = (int) (xOnScreen/scale);
      this.y = (int) (yOnScreen/scale);
      this.sizeOnScreen = (int) (SIZE*scale);
      this.selected = false;
      this.connectingToRooms = new Pair[6]; //rooms we can go to trough each of the 6 possible doors
      for (byte door = 1 ; door <= 6; ++door)
      {
         connectingToRooms[door-1] = null;
      }
      mayComeFromRooms = new LinkedList();
  }

  public Pair[] getConnectingToRooms() {return connectingToRooms;}
  public Room getConnectingToRoom(byte door) {return (Room) connectingToRooms[door].getFirst();}
  public LinkedList getMayComeFromRooms() {return mayComeFromRooms;}

  public boolean isDoorFree(byte door)
  {
     if ((door<0) || (door>6)) return false;
     return ((door == DOORNONE) || (connectingToRooms[door] == null));
  }

  public Wire lookForConnectionToRoom(Room room)
  {
     Pair pair = null; // pair<room, wire>
     for(byte door = 1; door <=6; ++door)
     {
        pair = connectingToRooms[door-1];
        if ((pair != null) && ((Room) pair.getFirst() == room)) return (Wire) pair.getSecond();
     }
     return null;
  }
  public byte lookForDoorConnectingToRoom(Room room)
  {
     Pair pair = null; // pair<room, wire> or pair<vnum, null>
     for(byte door = 0; door <6; ++door)
     {
        pair = connectingToRooms[door];
        if (pair != null)
        {
          try{if ((Room) pair.getFirst() == room) return door;}
          catch(Exception e){}
          try{if (((Integer) pair.getFirst()).intValue() == room.anatoliaRoom.getVnum()) return door;}
          catch(Exception e){}
        }
     }
     return DOORNONE;
  }


  public static void connect(Room fromRoom, byte fromDoor,
                             Room toRoom, byte toDoor, Wire byWire)
  {
     if (fromDoor != DOORNONE)
     {
        fromRoom.connectingToRooms[fromDoor] = new Pair(toRoom, byWire);
        toRoom.mayComeFromRooms.add(new Pair(fromRoom, new Byte(fromDoor)));
        fromRoom.anatoliaRoom.doors[fromDoor] = new DoorExit();
     }
     if (toDoor != DOORNONE)
     {
        toRoom.connectingToRooms[toDoor] = new Pair(fromRoom, byWire);
        fromRoom.mayComeFromRooms.add(new Pair(toRoom, new Byte(toDoor)));
        toRoom.anatoliaRoom.doors[toDoor] = new DoorExit();
     }
  }

  public static void disconnect(Room fromRoom, byte fromDoor,
                                Room toRoom, byte toDoor, Wire byWire)
  {
     if (fromDoor != DOORNONE)
     {
        if (((Wire) fromRoom.connectingToRooms[fromDoor].getSecond())== byWire)
           fromRoom.connectingToRooms[fromDoor]=null;

        toRoom.mayComeFromRooms.remove(new Pair(fromRoom, new Byte(fromDoor)));
        fromRoom.anatoliaRoom.doors[fromDoor] = null;
     }
     if (toDoor != DOORNONE)
     {
        toRoom.connectingToRooms[toDoor] = null;
        fromRoom.mayComeFromRooms.remove(new Pair(toRoom, new Byte(toDoor)));
        toRoom.anatoliaRoom.doors[toDoor] = null;
     }
  }
  public int getSizeOnScreen() {return sizeOnScreen;}
  public int getX() {return x;}
  public int getY() {return y;}
  public int getXOnScreen() {return xOnScreen;}
  public int getYOnScreen() {return yOnScreen;}
  public int getCandidateXOnScreen() {return candidateXOnScreen;}
  public int getCandidateYOnScreen() {return candidateYOnScreen;}

  public Room pseudoClone() //do not copy links, and change VNUM (anatolia property)
  {
     Room room = new Room(areaMap, xOnScreen, yOnScreen, scale, myCollection);
     room.anatoliaRoom = this.anatoliaRoom.pseudoClone(room);
     //don't forget to set anatoliaRoom;
     room.layer = this.layer;
     return room;
  }

  public boolean hasMoved()
  {
     return (    (candidateXOnScreen != xOnScreen)
              || (candidateYOnScreen != yOnScreen));
  }

  public void reScale(float newScale)
  {
      candidateXOnScreen = xOnScreen = (int) (x*newScale);
      candidateYOnScreen = yOnScreen = (int) (y*newScale);
      sizeOnScreen = (int) (SIZE*newScale);
      scale = newScale;
  }

  public void setPositionOnScreen(int newXOnScreen, int newYOnScreen)
  {
      x = (int) (newXOnScreen/scale);
      y = (int) (newYOnScreen/scale);
      candidateXOnScreen = xOnScreen = newXOnScreen;
      candidateYOnScreen = yOnScreen = newYOnScreen;
  }
  public void setLocation(int x, int y)
  {
      this.x = x;
      this.y = y;
      candidateXOnScreen = xOnScreen =  (int) (x*scale);
      candidateYOnScreen = yOnScreen =  (int) (y*scale);
  }

  public void translating(int deltaCandidateXOnScreen, int deltaCandidateYOnScreen)
  {
     this.candidateXOnScreen += deltaCandidateXOnScreen;
     this.candidateYOnScreen += deltaCandidateYOnScreen;
  }

  public void translate(int deltaXOnScreen, int deltaYOnScreen)
  {
      setPositionOnScreen(getXOnScreen()+deltaXOnScreen, getYOnScreen()+deltaYOnScreen);
  }


  public void paint(Graphics g)
  {
      int roomSizeOnScreen = (int) (SIZE*scale);
      int x2 = candidateXOnScreen + roomSizeOnScreen;
      int y2 = candidateYOnScreen + roomSizeOnScreen;

      if (anatoliaRoom.isExtern())
      {
         int radius = roomSizeOnScreen/2;
         g.fillOval(candidateXOnScreen+1, candidateYOnScreen+1,
                    roomSizeOnScreen-2, roomSizeOnScreen-2);
      }
      else
      {
         switch(anatoliaRoom.sectorType)
         {
            case AnatoliaConsts.ROOM_SECTOR_AIR       : g.setColor(Color.white)          ;break;
            case AnatoliaConsts.ROOM_SECTOR_CITY      : g.setColor(new Color(131,80,35)) ;break;
            case AnatoliaConsts.ROOM_SECTOR_INSIDE    : g.setColor(new Color(131,80,35)) ;break;
            case AnatoliaConsts.ROOM_SECTOR_FIELD     : g.setColor(new Color(0,250,0))   ;break;
            case AnatoliaConsts.ROOM_SECTOR_FOREST    : g.setColor(new Color(0,150,0))   ;break;
            case AnatoliaConsts.ROOM_SECTOR_HILL      : g.setColor(new Color(0,200,0))   ;break;
            case AnatoliaConsts.ROOM_SECTOR_MOUNTAIN  : g.setColor(Color.darkGray);break;
            case AnatoliaConsts.ROOM_SECTOR_WATER     : g.setColor(Color.cyan)     ;break;
            case AnatoliaConsts.ROOM_SECTOR_DEEPWATER : g.setColor(Color.blue)     ;break;
            case AnatoliaConsts.ROOM_SECTOR_DESERT    : g.setColor(Color.yellow)   ;break;
            default : g.setColor(Color.black);break;
         }
         g.fillRect(candidateXOnScreen+1                   , candidateYOnScreen+1,
                    roomSizeOnScreen-2                     , roomSizeOnScreen-2);
         g.setColor(Color.black);
      }

      g.fillOval(candidateXOnScreen+roomSizeOnScreen/2-roomSizeOnScreen/4,
                 candidateYOnScreen+roomSizeOnScreen/2-roomSizeOnScreen/4,
                 roomSizeOnScreen/2                    , roomSizeOnScreen/2);

      g.drawLine(candidateXOnScreen-5                 , candidateYOnScreen+roomSizeOnScreen/2,
                 x2+5                                 , candidateYOnScreen+roomSizeOnScreen/2);

      g.drawLine(candidateXOnScreen+roomSizeOnScreen/2 , candidateYOnScreen-5,
                 candidateXOnScreen+roomSizeOnScreen/2 , y2+5);

      g.drawLine(candidateXOnScreen-2                  , candidateYOnScreen-2,
                 x2+2                                  , y2+2);

      if (selected)
      {
         g.setColor(Color.red);
         g.drawRect(candidateXOnScreen-5, candidateYOnScreen-5,
                    sizeOnScreen+2*5    , sizeOnScreen+2*5);
         g.setColor(Color.black);
      }
  }

  public byte getDoor(int xMouse, int yMouse)
  {
      int centerX = candidateXOnScreen + sizeOnScreen/2;
      int centerY = candidateYOnScreen + sizeOnScreen/2;

      //Center
      if (Point2D.distance(centerX, centerY, xMouse, yMouse) < (sizeOnScreen/5))
        return DOORNONE;

      //NORTH
      Arc2D.Float arc = new Arc2D.Float(candidateXOnScreen-5, candidateYOnScreen-5,
                                        sizeOnScreen+10, sizeOnScreen+10,
                                        45, (45+45/2), Arc2D.PIE);
      if (arc.contains(xMouse, yMouse)) return DOORNORTH;
      //EAST
      arc.setAngleStart(-45/2);
      arc.setAngleExtent(45/2+45);
      if (arc.contains(xMouse, yMouse)) return DOOREAST;
      //SOUTH
      arc.setAngleStart(-135);
      arc.setAngleExtent(45+45/2);
      if (arc.contains(xMouse, yMouse)) return DOORSOUTH;
      //WEST
      arc.setAngleStart((135+180)/2);
      arc.setAngleExtent(45/2+45);
      if (arc.contains(xMouse, yMouse)) return DOORWEST;
      //UP
      arc.setAngleStart((90+135)/2);
      arc.setAngleExtent(45);
      if (arc.contains(xMouse, yMouse)) return DOORUP;
      //DOWN
      arc.setAngleStart((-90-45)/2);
      arc.setAngleExtent(45);
      if (arc.contains(xMouse, yMouse)) return DOORDOWN;

      return DOORNONE;
  }

  public Point getDoorPoint(byte door)
  {
      int centerX = candidateXOnScreen + sizeOnScreen/2;
      int centerY = candidateYOnScreen + sizeOnScreen/2;

      switch(door)
      {
         case DOORNONE  : return new Point(centerX, centerY);
         case DOORNORTH : return new Point(centerX,
                                           candidateYOnScreen-5);

         case DOOREAST  : return new Point(candidateXOnScreen+sizeOnScreen+5,
                                           centerY);

         case DOORSOUTH : return new Point(centerX,
                                           candidateYOnScreen+sizeOnScreen+5);

         case DOORWEST  : return new Point(candidateXOnScreen-5,
                                           centerY);

         case DOORUP    : return new Point(candidateXOnScreen-2,
                                           candidateYOnScreen-2);

         case DOORDOWN  : return new Point(candidateXOnScreen+sizeOnScreen+2,
                                           candidateYOnScreen+sizeOnScreen+2);
         default : return new Point(centerX, centerY);
      }
  }

  public void highlightDoor(Graphics g, int xMouse, int yMouse)
  {
      int centerX = xOnScreen + sizeOnScreen/2;
      int centerY = yOnScreen + sizeOnScreen/2;

      paint(g);
      g.setColor(Color.red);

      byte door = getDoor(xMouse, yMouse);
      switch(door)
      {
         case DOORNONE :  g.fillOval(centerX-sizeOnScreen/4, centerY-sizeOnScreen/4, sizeOnScreen/2, sizeOnScreen/2);
                          break;
         case DOORNORTH : g.drawLine(centerX, centerY, centerX, yOnScreen-5);
                          break;
         case DOOREAST  : g.drawLine(centerX, centerY, xOnScreen+sizeOnScreen+5, centerY);
                          break;
         case DOORSOUTH : g.drawLine(centerX, centerY, centerX, yOnScreen+sizeOnScreen+5);
                          break;
         case DOORWEST  : g.drawLine(centerX, centerY, xOnScreen-5, centerY);
                          break;
         case DOORUP    : g.drawLine(centerX, centerY, xOnScreen-2, yOnScreen-2);
                          break;
         case DOORDOWN  : g.drawLine(centerX, centerY, xOnScreen+sizeOnScreen+2, yOnScreen+sizeOnScreen+2);
                          break;
         default : break;
      }
      g.setColor(Color.black);
  }

  public boolean isContainedInRectangleOnScreen(Rectangle r)
  {
     return    areaMap.selectedLayers.contains(new Integer(this.layer))
            && r.contains(xOnScreen, yOnScreen, sizeOnScreen, sizeOnScreen);
  }

  public void writeObject(ObjectOutputStream output) throws IOException
  {
     output.writeLong(super.getObjectID());
     anatoliaRoom.writeObject(output);
     output.writeBoolean(autoVnum);
     output.writeInt(x);
     output.writeInt(y);
     output.writeInt(sizeOnScreen);
     output.writeInt(xOnScreen);
     output.writeInt(yOnScreen);
     output.writeInt(candidateXOnScreen);
     output.writeInt(candidateYOnScreen);
     output.writeInt(layer);
     output.writeFloat(scale);
     Pair pair;
     for(int door=1; door<=6 ; ++door)
     {
        pair = connectingToRooms[door-1]; //<room,wire>
        if (pair == null)
        {
           output.writeLong(0L);
           output.writeLong(0L);
        }
        else
        {
           output.writeLong(((Room) pair.getFirst()).getObjectID());
           output.writeLong(((Wire) pair.getSecond()).getObjectID());
        }
     }
  }

   public void readObject(ObjectInputStream input) throws IOException, ClassNotFoundException
   {
     super.setObjectID(input.readLong());
     anatoliaRoom.readObject(input);
     autoVnum           = input.readBoolean();
     x                  = input.readInt();
     y                  = input.readInt();
     sizeOnScreen       = input.readInt();
     xOnScreen          = input.readInt();
     yOnScreen          = input.readInt();
     candidateXOnScreen = input.readInt();
     candidateYOnScreen = input.readInt();
     layer              = input.readInt();
     scale              = input.readFloat();

     Pair pair = null;
     Pair nullPair = new Pair(new Long(0),new Long(0));
     for(int door=1; door<=6 ; ++door)
     {
        pair = new Pair(new Long(input.readLong()), new Long(input.readLong())); //<roomID, wireID>
        if (pair.equals(nullPair))
           connectingToRooms[door-1] = null; //no connection
        else
           connectingToRooms[door-1] = pair; //<roomID, wireID>
      }
   }

   public void establishLinks()
   {
     Pair pair = null;
     long theLongID = 0;
     for(int door=1; door<=6 ; ++door)
     {
        pair = connectingToRooms[door-1]; //<roomID,wireID>
        if (pair != null)
        {
           //!can be optimized
           theLongID = ((Long) pair.getFirst()).longValue();
           pair.setFirst(lookForObjectOfID(theLongID));
           theLongID = ((Long) pair.getSecond()).longValue();
           pair.setSecond(lookForObjectOfID(theLongID));
           connectingToRooms[door-1] = pair; //<room,wire>
        }
      }
   }

   public String toString()
   {
      return anatoliaRoom.toString();
   }
}