/*********************************************************************
** AreaBuilder is a program designed to build maps for MUDS like    **
** Anatolia.                                                        **
** Copyright (C) 2003 Pierre Chatelier                              **
**                                                                  **
** This program is free software; you can redistribute              **
** it and/or modify it under the terms of the GNU                   **
** General Public License as published by the                       **
** Free Software Foundation, version 2.                             **
**                                                                  **
** This program is distributed in the hope that it will be          **
** useful, but WITHOUT ANY WARRANTY; without even the implied       **
** warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. **
** See the GNU General Public License for more details.             **
**                                                                  **
** You should have received a copy of the                           **
** GNU General Public License along with this program; if not,      **
** write to the Free Software Foundation,                           **
** Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA      **
*********************************************************************/
package areabuilder;

import java.awt.*;
import java.util.*; //for LinkedList and HashMap
import java.io.*;   //for archivation

public class RoomCollection implements Serializable
{
  LinkedList roomList = new LinkedList();
  AnatoliaCollection anatoliaCollection = null; //for anatoliaRooms

  int xMinOnScreen = 0;
  int yMinOnScreen = 0;
  int xMaxOnScreen = 0;
  int yMaxOnScreen = 0;

  AreaMap areaMap = null;

  public RoomCollection(AreaMap areaMap)
  {
    this.areaMap = areaMap;
    this.anatoliaCollection = new AnatoliaCollection(areaMap);
  }

  public void clear()
  {
     roomList.clear();
     anatoliaCollection.clear();
     xMinOnScreen = 0;
     yMinOnScreen = 0;
     xMaxOnScreen = 0;
     yMaxOnScreen = 0;
  }

  public boolean isEmpty()
  {
    return roomList.isEmpty();
  }

  public int getXMinOnScreen() {return xMinOnScreen;}
  public int getYMinOnScreen() {return yMinOnScreen;}
  public int getXMaxOnScreen() {return xMaxOnScreen;}
  public int getYMaxOnScreen() {return yMaxOnScreen;}
  public Iterator getIterator() {return roomList.listIterator();}

  public void resetMinMax()
  {
    if (roomList.isEmpty())
    {
       xMinOnScreen = xMaxOnScreen = 0;
       yMinOnScreen = yMaxOnScreen = 0;
    }
    else
    {
       Iterator it = roomList.listIterator();
       Room room = (Room) it.next();
       xMinOnScreen = xMaxOnScreen = room.getXOnScreen();
       yMinOnScreen = yMaxOnScreen = room.getYOnScreen();
       while (it.hasNext())
       {
          room = (Room) it.next();
          xMinOnScreen = Math.min(xMinOnScreen, room.getXOnScreen());
          yMinOnScreen = Math.min(yMinOnScreen, room.getYOnScreen());
          xMaxOnScreen = Math.max(xMaxOnScreen, room.getXOnScreen());
          yMaxOnScreen = Math.max(yMaxOnScreen, room.getYOnScreen());
       }
    }
  }

  public void translate(int deltaXOnScreen, int deltaYOnScreen)
  {
    Iterator it = roomList.listIterator();
    Room room = null;
    while (it.hasNext())
    {
      room = (Room) it.next();
      room.translate(deltaXOnScreen, deltaYOnScreen);
    }
    xMinOnScreen += deltaXOnScreen;
    xMaxOnScreen += deltaXOnScreen;
    yMinOnScreen += deltaYOnScreen;
    yMaxOnScreen += deltaYOnScreen;
  }

  public void add(Room room)
  {
    if (roomList.isEmpty())
    {
       xMinOnScreen = xMaxOnScreen = room.getXOnScreen();
       yMinOnScreen = yMaxOnScreen = room.getYOnScreen();
    }
    else
    {
       xMinOnScreen = Math.min(xMinOnScreen, room.getXOnScreen());
       yMinOnScreen = Math.min(yMinOnScreen, room.getYOnScreen());
       xMaxOnScreen = Math.max(xMaxOnScreen, room.getXOnScreen());
       yMaxOnScreen = Math.max(yMaxOnScreen, room.getYOnScreen());
    }
    roomList.add(room);
    anatoliaCollection.add(room.anatoliaRoom);
    areaMap.objectTree.addNode(room.anatoliaRoom);
  }

  //returns the rank of the node in the tree
  public void remove(Room room)
  {
    roomList.remove(room);
    anatoliaCollection.remove(room.anatoliaRoom);
    resetMinMax();
  }

  public void paint(Graphics g)
  {
    Iterator it = roomList.listIterator();
    Room room = null;
    while (it.hasNext())
    {
      room = (Room) it.next();
      if (areaMap.selectedLayers.contains(new Integer(room.layer)))
      {
        room.paint(g);
      }
    }
  }

  public void reScale(float newScale)
  {
    Iterator it = roomList.listIterator();
    Room room = null;
    while (it.hasNext())
    {
      room = (Room) it.next();
      room.reScale(newScale);
    }
  }

  public Room inWhichRoom(int xMouse, int yMouse, float scale)
  {
    if (roomList.size() == 0) return null;
    int screenRoomSize = (int) (Room.SIZE*scale);

    ListIterator it = roomList.listIterator(roomList.size());
    Room room = null;
    while (it.hasPrevious())
    {
      room = (Room) it.previous();
      if (
            (areaMap.selectedLayers.contains(new Integer(room.layer)))
          &&(xMouse >= room.getXOnScreen())
          &&(xMouse <= room.getXOnScreen()+screenRoomSize)
          &&(yMouse >= room.getYOnScreen())
          &&(yMouse <= room.getYOnScreen()+screenRoomSize))
            return room;
    }
    return null;
  }

  public void buildRoomDescriptions(FileWriter output) throws IOException
  {
      Iterator it = roomList.iterator();
      while(it.hasNext())
      {
         ((Room) it.next()).anatoliaRoom.writeDescription(output);
      }
  }
  public void buildRoomResets(FileWriter output) throws IOException
  {
      Iterator it = roomList.iterator();
      while(it.hasNext())
      {
         ((Room) it.next()).anatoliaRoom.writeReset(output);
      }
  }

  public void writeObject(ObjectOutputStream output) throws IOException
  {
      anatoliaCollection.writeObject(output);
      output.writeInt(xMinOnScreen);
      output.writeInt(yMinOnScreen);
      output.writeInt(xMaxOnScreen);
      output.writeInt(yMaxOnScreen);
      output.writeInt(roomList.size());
      Iterator it = roomList.iterator();
      while(it.hasNext())
      {
         ((Room) it.next()).writeObject(output);
      }
  }

   public void readObject(ObjectInputStream input) throws IOException, ClassNotFoundException
   {
      this.clear();

      anatoliaCollection.readObject(input);
      xMinOnScreen = input.readInt();
      yMaxOnScreen = input.readInt();
      xMinOnScreen = input.readInt();
      yMaxOnScreen = input.readInt();
      int sizeOfRoomList = input.readInt();
      Room room = null;
      for(int i=0; i<sizeOfRoomList; ++i)
      {
         room = new Room(areaMap, anatoliaCollection);
         room.readObject(input);
         this.add(room);
      }
   }

   public void changeAreaNumber(int oldAreaNumber, int newAreaNumber)
   {
     anatoliaCollection.changeAreaNumber(oldAreaNumber, newAreaNumber);
   }

   public void establishLinks()
   {
      Iterator it = roomList.listIterator();
      Room room = null;
      while (it.hasNext())
      {
        room = (Room) it.next();
        room.establishLinks();
      }
  }
}